"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const moment_1 = __importDefault(require("moment"));
const testrail_js_api_1 = require("testrail-js-api");
const Status = {
    Passed: {
        value: testrail_js_api_1.TestStatus.Passed,
        text: "PASSED",
        color: "yellow",
    },
    Blocked: {
        value: testrail_js_api_1.TestStatus.Blocked,
        text: "SKIPPED",
        color: "green",
    },
    Failed: {
        value: testrail_js_api_1.TestStatus.Failed,
        text: "FAILED",
        color: "red",
    },
};
const loadJSON = (file) => {
    try {
        const data = fs_1.default.readFileSync(file, { encoding: "utf8" });
        if (data) {
            return JSON.parse(data);
        }
    }
    catch (error) {
    }
    return {};
};
const prepareConfig = (options = {}) => {
    const config = Object.assign(loadJSON(path_1.default.join(process.cwd(), ".testrailrc")), options);
    return {
        enabled: process.env.TESTRAIL_ENABLED === "true" || config.enabled || false,
        host: process.env.TESTRAIL_HOST || config.host,
        user: process.env.TESTRAIL_USER || config.user,
        apiKey: process.env.TESTRAIL_API_KEY || config.apiKey,
        projectId: Number((process.env.TESTRAIL_PROJECT_ID || config.projectId || "")
            .replace("P", "")
            .trim()),
        suiteId: Number((process.env.TESTRAIL_SUITE_ID || config.suiteId || "")
            .replace("S", "")
            .trim()),
        runId: Number((process.env.TESTRAIL_RUN_ID || config.runId || "")
            .replace("R", "")
            .trim()),
        runName: process.env.TESTRAIL_RUN_NAME ||
            config.runName ||
            "%BRANCH%#%BUILD% - %DATE%",
        runDescription: process.env.TESTRAIL_RUN_DESCRIPTION || config.runDescription,
        reference: process.env.TESTRAIL_REFERENCE || config.reference,
        branchEnv: process.env.TESTRAIL_BRANCH_ENV || config.branchEnv || "BRANCH",
        buildNoEnv: process.env.TESTRAIL_BUILD_NO_ENV || config.buildNoEnv || "BUILD_NUMBER",
        dateFormat: process.env.TESTRAIL_DATE_FORMAT ||
            config.dateFormat ||
            "YYYY-MM-DD HH:mm:ss",
        caseMeta: process.env.TESTRAIL_CASE_META || config.caseMeta || "CID",
        runCloseAfterDays: Number(process.env.TESTRAIL_RUN_CLOSE_AFTER_DAYS || config.runCloseAfterDays) || 0,
        uploadScreenshots: process.env.TESTRAIL_UPLOAD_SCREENSHOTS == "true" ||
            config.uploadScreenshots ||
            false,
        updateRunTestCases: process.env.TESTRAIL_UPDATE_RUN_TEST_CASES == "true" ||
            config.updateRunTestCases !== false,
    };
};
const prepareReportName = (config, branch, buildNo, userAgents) => {
    const date = moment_1.default().format(config.dateFormat);
    return config.runName
        .replace("%BRANCH%", branch)
        .replace("%BUILD%", buildNo)
        .replace("%DATE%", date)
        .replace("%AGENTS%", `(${userAgents.join(", ")})`);
};
const prepareRun = async (testrailAPI, config, runName, refs, caseIdList) => {
    const { projectId, suiteId, runDescription, runId, updateRunTestCases, } = config;
    let existingRun;
    if (runId) {
        const { value: returnedRun } = await throwOnApiError(testrailAPI.getRun(runId));
        existingRun = returnedRun;
    }
    else {
        const { value: runs } = await throwOnApiError(testrailAPI.getRuns(projectId, { is_completed: 0 }));
        existingRun = runs === null || runs === void 0 ? void 0 : runs.find((run) => run.refs === refs);
    }
    if (!updateRunTestCases) {
        if (existingRun) {
            return existingRun;
        }
        else {
            throw new Error(`[TestRail] Flag 'updateRunTestCases' enabled but the run was not found, please create it`);
        }
    }
    else if (existingRun) {
        const { value: tests } = await throwOnApiError(testrailAPI.getTests(existingRun.id));
        const currentCaseIds = (tests === null || tests === void 0 ? void 0 : tests.map((test) => test.case_id)) || [];
        const additionalDescription = "\n" + runDescription;
        const newDescription = existingRun.description
            ? existingRun.description.replace(additionalDescription, "") +
                additionalDescription
            : additionalDescription;
        await throwOnApiError(testrailAPI.updateRun(existingRun.id, {
            description: newDescription,
            case_ids: [...currentCaseIds, ...caseIdList],
        }));
        console.log(`[TestRail] Test run updated successfully: ${runName}`);
        return existingRun;
    }
    else {
        const payload = {
            suite_id: suiteId,
            include_all: false,
            case_ids: caseIdList,
            name: runName,
            description: runDescription,
            refs,
        };
        const { value: newRun } = await throwOnApiError(testrailAPI.addRun(projectId, payload));
        console.log(`[TestRail] Test run added successfully: ${runName}`);
        return newRun;
    }
};
const prepareReference = (config, branch, buildNo) => {
    return config.reference
        ? config.reference.replace("%BRANCH%", branch).replace("%BUILD%", buildNo)
        : "";
};
const verifyConfig = (config) => {
    const { enabled, host, user, apiKey, projectId, suiteId } = config;
    if (enabled) {
        if (!host) {
            console.log("[TestRail] Hostname was not provided.");
        }
        if (!user || !apiKey) {
            console.log("[TestRail] Username or api key was not provided.");
        }
        if (!projectId) {
            console.log("[TestRail] Project id was not provided.");
        }
        if (!suiteId) {
            console.log("[TestRail] Suite id was not provided.");
        }
        if (host && user && apiKey && projectId && suiteId) {
            return true;
        }
    }
    return false;
};
const throwOnApiError = async (apiResult) => {
    const { response, value } = await apiResult;
    if (response.status >= 400) {
        console.error("[TestRail] Error during API request");
        throw {
            url: response.url,
            status: response.status,
            message: value,
        };
    }
    return Promise.resolve({ response, value });
};
class TestcafeTestrailReporter {
    constructor() {
        this.reportTaskStart = async (_startTime, userAgents) => {
            this.userAgents = userAgents;
        };
        this.reportFixtureStart = async () => {
        };
        this.reportTestDone = async (name, testRunInfo, meta, formatError) => {
            const hasErr = testRunInfo.errs.length;
            let testStatus = null;
            if (testRunInfo.skipped) {
                testStatus = Status.Blocked;
            }
            else if (hasErr === 0) {
                testStatus = Status.Passed;
            }
            else {
                testStatus = Status.Failed;
            }
            let caseId = 0;
            if (meta[this.config.caseMeta]) {
                caseId = Number(meta[this.config.caseMeta].replace("C", "").trim());
            }
            if (caseId > 0) {
                const errorLog = testRunInfo.errs
                    .map((x) => {
                    const formatted = formatError(x).replace(/[\u001b\u009b][[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-ORZcf-nqry=><]/g, "");
                    return formatted;
                })
                    .join("\n");
                this.results.push({
                    case_id: caseId,
                    status_id: testStatus.value,
                    comment: `Test ${testStatus.text}\n${errorLog}`,
                });
                if (testRunInfo.screenshots.length) {
                    this.screenshots[caseId] = testRunInfo.screenshots;
                }
            }
            else {
                console.log(`[TestRail] Test missing the TestRail Case ID in test metadata: ${name}`);
            }
        };
        this.reportTaskDone = async (_endTime, _passed, _warnings, _result) => {
            const { host, user, apiKey, projectId, suiteId } = this.config;
            if (verifyConfig(this.config)) {
                try {
                    if (this.results.length) {
                        const runName = prepareReportName(this.config, this.branch, this.buildNo, this.userAgents);
                        const refs = prepareReference(this.config, this.branch, this.buildNo);
                        const caseIdList = this.results.map((result) => result.case_id);
                        const testrailAPI = new testrail_js_api_1.TestRail(host, user, apiKey);
                        const { value: caseList } = await throwOnApiError(testrailAPI.getCases(projectId, { suite_id: suiteId }));
                        const existingCaseIds = caseList.map((item) => item.id);
                        caseIdList.forEach((id) => {
                            if (!existingCaseIds.includes(id)) {
                                console.error(`[TestRail] All TestRail mappings should be valid. Following test case id does not exist in TestRail: ${id}.`);
                            }
                        });
                        const run = await prepareRun(testrailAPI, this.config, runName, refs, caseIdList);
                        await this.publishTestResults(testrailAPI, run, this.results);
                        await this.closeOldRuns(testrailAPI, this.config);
                    }
                    else {
                        console.log("[TestRail] No test case data found to publish");
                    }
                }
                catch (error) {
                    console.error("[TestRail] Sending report to TestRail failed", error);
                    throw error;
                }
            }
        };
        this.closeOldRuns = async (testrailAPI, config) => {
            if (config.runCloseAfterDays) {
                const { value: runs } = await throwOnApiError(testrailAPI.getRuns(config.projectId, { is_completed: 0 }));
                if (runs.length) {
                    for (let i = 0; i < runs.length; i++) {
                        const shouldClose = moment_1.default.unix(runs[i].created_on) <=
                            moment_1.default().subtract(config.runCloseAfterDays, "days");
                        if (shouldClose) {
                            console.log(`[TestRail] Closing test run ${runs[i].id}: ${runs[i].name}`);
                            await throwOnApiError(testrailAPI.closeRun(runs[i].id));
                        }
                    }
                }
            }
        };
        this.publishTestResults = async (testrailAPI, run, resultsToPush) => {
            const runId = run.id;
            const { value: results } = await throwOnApiError(testrailAPI.addResultsForCases(runId, resultsToPush));
            const { value: tests } = await throwOnApiError(testrailAPI.getTests(runId));
            if (this.config.uploadScreenshots) {
                console.log("[TestRail] Uploading screenshots...");
                for (let i = 0; i < resultsToPush.length; i++) {
                    const test = tests.find((test) => test.case_id === resultsToPush[i].case_id);
                    const result = results.find((result) => result.test_id === (test === null || test === void 0 ? void 0 : test.id));
                    if (result) {
                        const screenshots = this.screenshots[resultsToPush[i].case_id];
                        if (screenshots) {
                            for (let j = 0; j < screenshots.length; j++) {
                                await throwOnApiError(testrailAPI.addAttachmentToResult(result.id, screenshots[j].screenshotPath));
                            }
                        }
                    }
                    else {
                        console.error(`[TestRail] Could not upload screenshot for a failed test. Case ID: ${resultsToPush[i].caseId}. Test ID: ${test === null || test === void 0 ? void 0 : test.id}`);
                    }
                }
            }
            if (results.length == 0) {
                console.log("[TestRail] No data has been published to TestRail");
            }
            else {
                console.log("[TestRail] Test results added to the TestRail successfully");
            }
        };
        this.config = prepareConfig();
        this.branch = process.env[this.config.branchEnv] || "master";
        this.buildNo = process.env[this.config.buildNoEnv] || "unknown";
        this.noColors = false;
        this.results = [];
        this.screenshots = {};
    }
}
module.exports = () => {
    const reporter = new TestcafeTestrailReporter();
    return {
        reportTaskStart: reporter.reportTaskStart,
        reportFixtureStart: reporter.reportFixtureStart,
        async reportTestDone(name, testRunInfo, meta) {
            return reporter.reportTestDone(name, testRunInfo, meta, this.formatError.bind(this));
        },
        reportTaskDone: reporter.reportTaskDone,
        reporter,
    };
};
